@@ -58,6 +58,7 @@ class FileSystemModel(gtk.ListStore):
 	# Set in FileSystemModel::__load_icons()
 	file_icon = None
 	dir_icon = None
+	dir_unreadable_icon = None
 	video, audio, tap, tap_ini = ('.rec', '.mp3', '.tap', '.ini')
 	icon_dict = { video : None,
 	              audio : None,
@@ -107,6 +108,7 @@ class FileSystemModel(gtk.ListStore):
 
 		# List of icons and their icon theme name
 		icons = { 'dir_icon' : 'folder',
+		          'dir_unreadable_icon' : 'emblem-unreadable',
 		          'video_icon' : 'video-x-generic',
 		          'audio_icon' : 'audio-x-generic',
 		          'tap_icon' : 'application-x-executable',
@@ -122,6 +124,7 @@ class FileSystemModel(gtk.ListStore):
 				pass
 
 		FileSystemModel.dir_icon = icons['dir_icon']
+		FileSystemModel.dir_unreadable_icon = icons['dir_unreadable_icon']
 		FileSystemModel.icon_dict[FileSystemModel.video] = icons['video_icon']
 		FileSystemModel.icon_dict[FileSystemModel.audio] = icons['audio_icon']
 		FileSystemModel.icon_dict[FileSystemModel.tap] = icons['tap_icon']
@@ -132,6 +135,9 @@ class FileSystemModel(gtk.ListStore):
 			FileSystemModel.dir_icon = img.render_icon(gtk.STOCK_DIRECTORY,
 		                                               gtk.ICON_SIZE_LARGE_TOOLBAR)
 		
+		if FileSystemModel.dir_unreadable_icon == None:
+			FileSystemModel.dir_unreadable_icon = gtk.gdk.pixbuf_new_from_file(datadir + '/' + 'unreadable.png')
+			
 		if FileSystemModel.icon_dict[FileSystemModel.video] == None:	
 			FileSystemModel.icon_dict[FileSystemModel.video] = gtk.gdk.pixbuf_new_from_file(datadir + '/' + 'video.png')
 
@@ -181,6 +187,9 @@ class FileSystemModel(gtk.ListStore):
 		FileSystemModel.__connect_theme_change_callback(self.changeDir)
 
 	def abspath(self, file):
+		if file[0] == self.slash:
+			return file
+		
 		if self.current_dir[-1] != self.slash and file[0] != self.slash:
 			return self.current_dir + self.slash + file
 		else:
@@ -336,8 +345,7 @@ class PCFileSystemModel(FileSystemModel):
 	def changeDir(self, dir=None):
 		if dir:
 			# Append CWD if dir is not an absolute path
-			if dir[0] != self.slash:
-				dir = self.abspath(dir)
+			dir = self.abspath(dir)
 		else:
 			dir = self.current_dir
 
@@ -394,8 +402,7 @@ class PCFileSystemModel(FileSystemModel):
 			os.remove(path)
 		
 	def exists(self, file):
-		if file[0] != self.slash:
-			file = self.abspath(file)
+		file = self.abspath(file)
 		
 		return os.access(file, os.F_OK)
 		
@@ -462,8 +469,7 @@ class PVRFileSystemModel(FileSystemModel):
 		"""
 		if dir:
 			# Append CWD if dir is not an absolute path
-			if dir[0] != self.slash:
-				dir = self.abspath(dir)
+			dir = self.abspath(dir)
 		else:
 			dir = self.current_dir
 
@@ -480,6 +486,9 @@ class PVRFileSystemModel(FileSystemModel):
 		if dir_node == None:
 			return
 		
+		if not dir_node.readable:
+			raise PVRFileSystemError('%s is unreadable.' % (dir))
+			
 		self.current_dir = dir
 		
 		# Clear model
@@ -496,7 +505,10 @@ class PVRFileSystemModel(FileSystemModel):
 			self.append(file_info)
 			
 		for dir, dir_info in dir_node.getDirectories():
-			dir_info.insert(FileSystemModel.ICON_COL, FileSystemModel.dir_icon)
+			if dir.readable == False:
+				dir_info.insert(FileSystemModel.ICON_COL, FileSystemModel.dir_unreadable_icon)
+			else:
+				dir_info.insert(FileSystemModel.ICON_COL, FileSystemModel.dir_icon)
 			
 			dir_info[FileSystemModel.SIZE_COL] = humanReadableSize(int(dir_info[FileSystemModel.SIZE_COL]))
 			
@@ -585,57 +597,67 @@ class PVRFileSystemModel(FileSystemModel):
 		
 		self.puppy.rename(self.abspath(old), self.abspath(new))
 		
-	def scanDirectory(self, dir):
+	def scanDirectory(self, dir, exclude=[]):
 		"""Recursively scan a directory for all subdirectories and files.
 		
 		This function recursivley scans the given directory for all subdirectories
 		and files. Each subdirectory and file is stored in the DirectoryNode object.
+		
+		exclude    List of directories to exclude from scan.
+				
 		Return: DirectoryNode object representing the directory
 		"""
+		
 		dir_node = DirectoryNode(dir.split('\\')[-1])
-		pvr_files = self.puppy.listDir(dir)
+		
+		if dir in exclude:
+			return dir_node
+		
+		try:
+			pvr_files = self.puppy.listDir(dir)
+		except PuppyBusyError:
+			# Raise PuppyBusyError so we know when failed to scan a directory
+			# because puppy was busy, probably due to a transfer.
+			raise
+		except PuppyError:
+			# Mark directory as unreadable
+			dir_node.readable = False
+			
+			# Don't raise exception so we can continue scanning.
+			return dir_node
 		
 		for file in pvr_files:
 			if file[0] == 'd':
 				if file[1] != '..':
-					dir_node.addDirectory(self.scanDirectory(dir + '\\' + file[1]), file)
+					dir_node.addDirectory(self.scanDirectory(dir + '\\' + file[1], exclude), file)
 			else:
 				dir_node.addFile(file[1], file)
 
 		return dir_node
 	
-	def updateCache(self):
+	def updateCache(self, exclude=[]):
 		"""Update the PVR file system cache.
 		
 		This function scans the entire PVR file system and creates a tree
 		representing all the files and directories.
+		
+		exclude    List of directories to exclude from cache.
+		
+		Returns:   True if cache successfully updated.
+		           False if failed to update cache.
 		"""
 		if DEBUG:
 			print 'Start updateCache()'
-			
+
+		new_dir_tree = self.scanDirectory('', exclude)
+
+		# Couldn't even scan root directory on PVR. Return False.
+		if new_dir_tree.readable == False:
+			return False
+		
 		debugLocking('locking')
 		self.dir_tree_lock.acquire()
 
-		new_dir_tree = None
-
-		# Attempt to update cache twice in case puppy was busy the first time
-		for i in xrange(2):
-			try:
-				new_dir_tree = self.scanDirectory('')
-			except PuppyBusyError:
-				# Sleep for 1 second before trying again
-				time.sleep(1)
-				continue
-			except PuppyError:
-				debugLocking('unlocking')
-				self.dir_tree_lock.release()
-				raise
-			break
-
-		# Failed to update cache. Use existing cache.
-		if new_dir_tree == None:
-			new_dir_tree = self.dir_tree
-			
 		self.dir_tree = new_dir_tree
 
 		debugLocking('unlocking')
@@ -643,8 +665,10 @@ class PVRFileSystemModel(FileSystemModel):
 		
 		if DEBUG:
 			print 'End updateCache()'
+			
+		return True
 
-	def updateDirectory(self, dir):
+	def updateDirectory(self, dir, exclude=[]):
 		"""Update the directory node in the directory tree.
 		
 		"""
@@ -653,53 +677,38 @@ class PVRFileSystemModel(FileSystemModel):
 			
 		# Strip trailing slash
 		dir = dir.rstrip(self.slash)
-
-		debugLocking('locking')
-		self.dir_tree_lock.acquire()
-
-		# Call scanDirectory() directly if dir is root dir, i.e. '\'
+		
+		# Just call updateCache() if dir is root dir
 		if len(dir) == 0:
-			self.dir_tree = self.scanDirectory('')
-			debugLocking('unlocking')
-			self.dir_tree_lock.release()
-			if DEBUG:
-				print 'End updateDirectory(\\)'
-			return
+			return self.updateCache(exclude)
+
+		# Remove dir from exclusion list before scanning dir.
+		dir = self.abspath(dir)
+		new_exclude = copy.copy(exclude)
+		if dir in exclude:
+			new_exclude.remove(dir)
+		
+		new_node = self.scanDirectory(dir, new_exclude)
 
 		# Get components of directory path as a list
 		path = dir.split('\\')
 
+		debugLocking('locking')
+		self.dir_tree_lock.acquire()
+
 		# Find parent node		
 		parent_node = self.dir_tree
 		for name in path[:-1]:
 			if len(name) == 0:
 				continue
-			par_node, node_info = parent_node.getDirectory(name)
+			parent_node, node_info = parent_node.getDirectory(name)
 			if parent_node == None:
 				break
-				
+
+		# Get node_info for existing node
 		cur_node, node_info = parent_node.getDirectory(path[-1])
 		
-		new_node = None
-		# Attempt to update cache twice in case puppy was busy the first time
-		for i in xrange(2):
-			try:
-				new_node = self.scanDirectory(dir)
-			except PuppyBusyError:
-				# Sleep for 1 second before trying again
-				time.sleep(1)
-				continue
-			except PuppyError:
-				debugLocking('unlocking')
-				self.dir_tree_lock.release()
-				raise
-			break
-
-		# Failed to update cache. Use existing node.
-		if new_node == None:
-			new_node = cur_node
-
-		# Replace wit new node
+		# Replace with new node
 		parent_node.addDirectory(new_node, node_info)
 		
 		debugLocking('unlocking')
@@ -714,6 +723,7 @@ class DirectoryNode:
 		self.name = name
 		self.sub_directories = {}
 		self.files = []
+		self.readable = True
 		
 	def addDirectory(self, dir, info):
 		if DEBUG:
@@ -759,3 +769,9 @@ class DirectoryNode:
 			
 		for file in self.files:
 			print file
+
+class PVRFileSystemError(Exception):
+	def __init__(self, value):
+		self.value = value
+	def __str__(self):
+		return self.value
